﻿Public Class Form1

    Sub RefreshLists()
        
        'fill GUI lists
        lvExportTable.BeginUpdate()
        lvExportTable.Items.Clear()
        For Each i In pack.ExportTable
            Dim item = lvExportTable.Items.Add(i.Name)
            item.SubItems.AddRange({TypeToName(i.TypeNum), i.HeaderSize, i.HeaderPtr, i.DataPtr, i.DataSize})
        Next
        lvExportTable.EndUpdate()
        lbNameTable.BeginUpdate()
        lbNameTable.Items.Clear()
        For Each i In pack.NameTable
            lbNameTable.Items.Add(i)
        Next
        lbNameTable.EndUpdate()

    End Sub

    


    Private Sub btnCopyDataAddr_Click(sender As System.Object, e As System.EventArgs) Handles btnCopyDataAddr.Click
        If lvExportTable.SelectedItems.Count = 1 Then
            My.Computer.Clipboard.SetText(lvExportTable.SelectedItems(0).SubItems(4).Text)
        End If

    End Sub

    Private Sub btnCopyHeaderAddr_Click(sender As System.Object, e As System.EventArgs) Handles btnCopyHeaderAddr.Click
        If lvExportTable.SelectedItems.Count = 1 Then
            My.Computer.Clipboard.SetText(lvExportTable.SelectedItems(0).SubItems(3).Text)
        End If
    End Sub

    Private Sub btnPreview_Click(sender As System.Object, e As System.EventArgs) Handles btnPreview.Click
        If lvExportTable.SelectedItems.Count <> 1 Then Return

        Dim selectedItem As ExportItem = pack.ExportTable(lvExportTable.SelectedIndices(0))

        Select Case selectedItem.TypeNum
            Case EResType.RES_TextBuffer
                MessageBox.Show(selectedItem.GetText)
            Case EResType.RES_Sound
                My.Computer.Audio.Play(selectedItem.GetSound(), AudioPlayMode.Background)
            Case EResType.RES_Texture
                Dim texInfo = selectedItem.GetTextureInfo()

                MessageBox.Show("Palette num: " & texInfo.PaletteNum & vbCrLf &
                                "Resolution: " & texInfo.Width & "x" & texInfo.Height & vbCrLf &
                                "Has mipmaps: " & texInfo.HasMipmaps)

                TexPreviewWindow.Show()
                TexPreviewWindow.PictureBox1.Image = selectedItem.GetTexture()

            Case EResType.RES_Polys
                MessageBox.Show(selectedItem.GetModel)
            Case EResType.RES_Class
                Dim frm As New ClassPreviewForm
                frm.SetForClass(selectedItem)
                frm.Show()
            Case Else
                MessageBox.Show("Type not supported for preview.")
        End Select
    End Sub

    Private Sub lvExportTable_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles lvExportTable.SelectedIndexChanged
        If lvExportTable.SelectedItems.Count = 1 Then
            lbResourceNum.Text = lvExportTable.SelectedIndices(0) + 1
        Else
            lbResourceNum.Text = ""
        End If

    End Sub

    Private Sub btnSaveBinData_Click(sender As System.Object, e As System.EventArgs) Handles btnSaveBinData.Click
        If lvExportTable.SelectedItems.Count <> 1 Then Return
        Dim selectedItem As ExportItem = pack.ExportTable(lvExportTable.SelectedIndices(0))

        SaveBinDlg.FileName = selectedItem.Name
        If SaveBinDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            IO.File.WriteAllBytes(SaveBinDlg.FileName, selectedItem.GetData())
        End If
    End Sub

    Private Sub btnSaveBinHeader_Click(sender As System.Object, e As System.EventArgs) Handles btnSaveBinHeader.Click
        If lvExportTable.SelectedItems.Count <> 1 Then Return
        Dim selectedItem As ExportItem = pack.ExportTable(lvExportTable.SelectedIndices(0))

        SaveBinDlg.FileName = selectedItem.Name & " header"
        If SaveBinDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            IO.File.WriteAllBytes(SaveBinDlg.FileName, selectedItem.GetHeader())
        End If
    End Sub

    Private Sub btnExport_Click(sender As System.Object, e As System.EventArgs) Handles btnExport.Click
        If FolderDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            For i = 0 To pack.ExportTable.Count - 1
                Dim item As ExportItem = pack.ExportTable(i)

                If Exportable(i) Then
                    Try
                        ExportResource(i, IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ExportResourceExt(i)))
                    Catch ex As Exception
                        MessageBox.Show("Exporting " & i & " (" & pack.ExportTable(i).Name & ") failed." & vbCrLf & vbCrLf & "Error type:" & vbCrLf & ex.GetType.Name & vbCrLf & "Details:" & ex.Message & vbCrLf & "Stack:" & ex.StackTrace, "Export error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                    End Try
                End If

                'Select Case item.TypeNum
                '    Case 4 'code
                '        r.BaseStream.Position = item.DataPtr
                '        IO.File.WriteAllBytes(IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ".txt"), r.ReadBytes(item.DataSize))
                '    Case 5 'texture
                '        ReadTexture(i).Save(IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ".png"), Imaging.ImageFormat.Png)
                '    Case 12 'sound
                '        r.BaseStream.Position = item.DataPtr + 252
                '        IO.File.WriteAllBytes(IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ".wav"), r.ReadBytes(item.DataSize - 252))
                'End Select

                ProgressBar1.Value = i / pack.ExportTable.Count * 100
                ProgressBar1.Refresh()
            Next

            MessageBox.Show("Finished.")
        End If
    End Sub

    ''' <summary>
    ''' Tries to export resource [num]. Warning: in case of mesh resources, two files will be produced, with filename endings "_d" and "_a".
    ''' </summary>
    ''' <returns>Returns whether exporting was possible.</returns>
    Function ExportResource(num As Integer, file As String) As Boolean
        Dim res As ExportItem = pack.ExportTable(num)

        Select Case res.TypeNum
            Case 4 'code
                IO.File.WriteAllBytes(file, res.GetData())
                Return True
            Case 5 'texture
                res.GetTexture().Save(file, Imaging.ImageFormat.Png)
                Return True
            Case 12 'sound
                IO.File.WriteAllBytes(file, res.GetSound())
                Return True
            Case 14 'mesh
                Dim dataFileName As String
                Dim animFileName As String
                Dim fileNameStart As String
                Dim fileNameEnd As String
                fileNameStart = IO.Path.Combine(IO.Path.GetDirectoryName(file), IO.Path.GetFileNameWithoutExtension(file))
                fileNameEnd = IO.Path.GetExtension(file)
                dataFileName = fileNameStart & "_d" & fileNameEnd
                animFileName = fileNameStart & "_a" & fileNameEnd

                res.SaveMesh(animFileName, dataFileName)

                Return True
            Case Else
                Return False
        End Select
    End Function

    ''' <summary>
    ''' Returns file extension for exported resource [num]. Returns "" if not exportable.
    ''' </summary>
    Function ExportResourceExt(num As Integer) As String
        Dim res As ExportItem = pack.ExportTable(num)

        Select Case res.TypeNum
            Case 4
                Return ".txt"
            Case 5
                Return ".png"
            Case 12
                Return ".wav"
            Case 14
                Return ".3d"
        End Select

        Return ""
    End Function

    ''' <summary>
    ''' Returns whether resource [num] can be exported.
    ''' </summary>
    Function Exportable(num As Integer) As Boolean
        Select Case pack.ExportTable(num).TypeNum
            Case 4, 5, 12, 14
                Return True
            Case Else
                Return False
        End Select
    End Function

    Private Sub btnOpen_Click(sender As System.Object, e As System.EventArgs) Handles btnOpen.Click
        If OpenDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            pack = New Package(OpenDlg.FileName)
            RefreshLists()
        End If
        
    End Sub

    Private Sub lbNameTable_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles lbNameTable.SelectedIndexChanged
        If lbNameTable.SelectedItems.Count = 1 Then
            lbResourceNum.Text = lbNameTable.SelectedIndex
        End If
    End Sub

    Private Sub btnExportT3D_Click(sender As System.Object, e As System.EventArgs) Handles btnExportT3D.Click
        ExportT3DForm.ShowDialog()
    End Sub

End Class
