﻿Public Module PackageClasses
    Public pack As Package

    Public Class Package

        Public r As UnBinaryReader

        Public FileHeaderPtr As UInt32
        Public PackageVersion As UInt32 'should be 9
        Public ExportCount As UInt32
        Public ExportTablePtr As UInt32
        Public ResHeadersPtr As UInt32
        Public NameCount As UInt32
        Public NameTablePtr As UInt32

        Public NameTable As New List(Of String)
        Public ExportTable As New List(Of ExportItem)

        ' These affect T3D export
        Public RemapPolyFlags As Boolean = True
        Public UseEnumNames As Boolean = True
        Public RemapProps As Boolean = True
        Public IncludeAllProps As Boolean = False
        Public FinalSyntax As Boolean = True
        Public ExportCameras As Boolean = False

        Public Indent As Byte = 0

        Function Tab() As String
            Return Space(Indent)
        End Function

        Sub New(file As String)
            r = New UnBinaryReader(file)

            'go to beginning of file header
            Me.FileHeaderPtr = r.BaseStream.Length - 164
            r.BaseStream.Position = Me.FileHeaderPtr

            'read file header
            Me.PackageVersion = r.ReadUInt32()
            Me.ExportCount = r.ReadUInt32()
            r.ReadUInt32()
            Me.NameCount = r.ReadUInt32()
            Me.NameTablePtr = r.ReadUInt32()
            r.ReadUInt32()
            Me.ExportTablePtr = r.ReadUInt32()
            Me.ResHeadersPtr = r.ReadUInt32()

            'read export table
            For i = 0 To Me.ExportCount - 1
                r.BaseStream.Position = Me.ExportTablePtr + 44 * i
                Dim curResName As String = Me.r.ReadCString(36)
                r.BaseStream.Position = Me.ExportTablePtr + 44 * i + 36
                Dim curResType As UInt32 = r.ReadUInt32
                Dim curResHeaderSize As UInt32 = r.ReadUInt32

                Me.ExportTable.Add(New ExportItem With {.MyLevel = Me, .Name = curResName, .TypeNum = curResType, .HeaderSize = curResHeaderSize})

            Next

            'calculate header pointers
            Dim sum As Integer = Me.ResHeadersPtr
            For Each i In Me.ExportTable
                i.HeaderPtr = sum
                sum += i.HeaderSize
            Next

            'read data pointers & lengths
            For Each i In Me.ExportTable
                r.BaseStream.Position = i.HeaderPtr + 60
                i.DataPtr = r.ReadUInt32()
                i.DataSize = r.ReadUInt32()
            Next

            'read name table
            For i = 0 To Me.NameCount - 1
                r.BaseStream.Position = Me.NameTablePtr + i * 36
                Me.NameTable.Add(Me.r.ReadCString(36))
            Next
        End Sub

        Protected Overrides Sub Finalize()
            r.Close()
        End Sub

        Sub ExportT3D(file As String)
            Dim w As New IO.StreamWriter(file)

            Dim brushList As Integer() = Nothing
            For i = 0 To ExportTable.Count - 1
                If ExportTable(i).TypeNum = EResType.RES_Array Then
                    brushList = ExportTable(i).GetBrushList
                    Exit For
                End If
            Next

            w.WriteLine("Begin Map" & If(FinalSyntax, "", " Brushes=" & brushList.GetLength(0)))

            Indent += 3

            If brushList IsNot Nothing Then
                For i = 0 To brushList.GetUpperBound(0)
                    w.Write(ExportTable(brushList(i)).GetBrush)
                Next
            Else
                MessageBox.Show("Brush list not found.")
            End If

            Dim actorList As ExportItem = Nothing
            For i = 0 To ExportTable.Count - 1
                If ExportTable(i).TypeNum = EResType.RES_ActorList Then
                    actorList = ExportTable(i)
                    Exit For
                End If
            Next

            If actorList IsNot Nothing Then
                Dim actorListElements As UInt16
                r.BaseStream.Position = actorList.HeaderPtr + 76
                actorListElements = r.ReadUInt16

                Dim ClassInfos As New List(Of ClassInfo)

                For i = 0 To ExportTable.Count - 1
                    If ExportTable(i).TypeNum = EResType.RES_Class Then
                        Dim clsInfo = New ClassInfo
                        clsInfo.Index = i + 1
                        clsInfo.Properties = ExportTable(i).GetClassProperties
                        clsInfo.Defaults = ExportTable(i).GetClassDefaults(clsInfo.Properties)
                        ClassInfos.Add(clsInfo)
                    End If
                Next

                If Not FinalSyntax Then
                    w.WriteLine(Tab() & "Begin ActorList")
                    Indent += 3
                End If

                For i = 0 To actorListElements - 1
                    r.BaseStream.Position = actorList.DataPtr + i * ActorSize()
                    Dim actClass = r.ReadInt32
                    If actClass = 0 Then Continue For

                    If (Not ExportCameras) And GetResName(actClass) = "Camera" Then
                        Continue For
                    End If

                    Dim curClassInfo As ClassInfo = Nothing
                    For j = 0 To ClassInfos.Count - 1
                        If ClassInfos(j).Index = actClass Then
                            curClassInfo = ClassInfos(j)
                            Exit For
                        End If
                    Next
                    If curClassInfo Is Nothing Then
                        MessageBox.Show("Bad class index")
                        Continue For
                    End If

                    Dim allProps = actorList.GetActor(i, curClassInfo.Properties)
                    Dim usedProps As List(Of ActorProperty)
                    If IncludeAllProps Then
                        usedProps = allProps
                    Else
                        usedProps = FilterProperties(allProps, curClassInfo.Defaults)
                    End If

                    If RemapProps Then
                        usedProps = TranslateProperties(usedProps)
                    End If

                    Dim myIndex = -1
                    If Not FinalSyntax Then
                        For j = 0 To usedProps.Count - 1
                            If usedProps(j).Name = "Me" Then
                                myIndex = usedProps(j).Value
                                Exit For
                            End If
                        Next
                    End If

                    w.WriteLine(Tab() & "Begin Actor" & If(FinalSyntax, "", " Index=" & myIndex) & " Class=" & GetResName(actClass))
                    Indent += 3
                    For j = 0 To usedProps.Count - 1
                        'Mover's model
                        If usedProps(j).Name = "Brush" Then
                            Dim model As ExportItem = Nothing
                            For k = 0 To ExportTable.Count - 1
                                If ExportTable(k).TypeNum = EResType.RES_Model AndAlso
                                   ExportTable(k).Name = usedProps(j).Value Then
                                    model = ExportTable(k)
                                    Exit For
                                End If
                            Next
                            If model Is Nothing Then
                                MessageBox.Show("Mover's model (" & usedProps(j).Value & ") not found.")
                            Else
                                If FinalSyntax Then
                                    w.Write(model.GetMissingMoverProps)
                                    w.Write(model.GetBrushModel.GetModel)
                                Else
                                    w.Write(model.GetBrush)
                                End If
                            End If

                        End If

                        w.WriteLine(Tab() & usedProps(j).Name & "=" & usedProps(j).Value)
                    Next
                    Indent -= 3
                    w.WriteLine(Tab() & "End Actor")
                Next

                If Not FinalSyntax Then
                    Indent -= 3
                    w.WriteLine(Tab() & "End ActorList")
                End If
            Else
                MessageBox.Show("Actor list not found.")
            End If

            Indent -= 3
            w.WriteLine("End Map")

            w.Close()
        End Sub

        Class ClassInfo
            Public Index As Integer 'from 1
            Public Properties As ClassProperty()
            Public Defaults As List(Of ActorProperty)
        End Class

        Shared Function FilterProperties(actual As List(Of ActorProperty), defaults As List(Of ActorProperty)) As List(Of ActorProperty)
            If actual.Count <> defaults.Count Then
                Throw New Exception("FilterProperties: Sizes don't match.")
            End If

            Dim result As New List(Of ActorProperty)

            For i = 0 To actual.Count - 1
                If actual(i).Name <> defaults(i).Name Then
                    Throw New Exception("FilterProperties: Names don't match.")
                Else
                    If actual(i).Value <> defaults(i).Value Then
                        result.Add(actual(i))
                    End If
                End If
            Next

            Return result
        End Function

        Shared Function TranslateProperties(props As List(Of ActorProperty)) As List(Of ActorProperty)
            Dim result As New List(Of ActorProperty)

            For i = 0 To props.Count - 1
                If ShouldIgnoreProperty(props(i).Name) Then
                    Continue For
                End If

                result.Add(RemapProperty(props(i)))
            Next

            Return result
        End Function

        Function GetResName(index As Integer) As String
            If index = 0 Then
                Return "None"
            ElseIf index < 0 Or index > ExportTable.Count Then
                Return ""
            Else 'If index > 0 Then
                Return ExportTable(index - 1).Name
            End If
        End Function

        Function GetName(num As UInt16) As String
            If PackageVersion <= 9 Then
                '0.83
                If num = 65535 Then
                    Return "None"
                ElseIf num < NameTable.Count Then
                    Return NameTable(num)
                Else
                    Return ""
                End If
            Else
                '0.84
                If num = 0 Then
                    Return "None"
                ElseIf num <= NameTable.Count Then
                    Return NameTable(num - 1)
                Else
                    Return ""
                End If
            End If


        End Function

        Function ReadActor(address As Integer, props As ClassProperty()) As List(Of ActorProperty)
            Dim result As New List(Of ActorProperty)

            For i = 0 To props.GetUpperBound(0)
                'r.BaseStream.Position = address + props(i).Offset

                Dim isArray = (((PackageVersion <= 9) And (props(i).ArraySize > 0)) Or
                               ((PackageVersion > 9) And (props(i).ArraySize > 1))) And
                                (props(i).Type <> EPropType.T_String)
                Dim elements = 1
                If isArray Then elements = props(i).ArraySize

                For j = 0 To elements - 1
                    r.BaseStream.Position = address + props(i).Offset + j * props(i).ElementSize

                    Dim prop As New ActorProperty

                    prop.Name = props(i).Name
                    If isArray Then prop.Name &= "(" & j & ")"

                    Select Case props(i).Type
                        Case EPropType.T_Int
                            prop.Value = r.ReadInt32
                        Case EPropType.T_Real
                            prop.Value = NumToStr(r.ReadSingle)
                        Case EPropType.T_Actor
                            'TODO
                            prop.Value = r.ReadUInt16
                        Case EPropType.T_Bool
                            prop.Value = CBool(r.ReadInt32 And props(i).Misc)
                        Case EPropType.T_Byte
                            Dim num = r.ReadByte

                            If UseEnumNames And props(i).Misc <> 0 Then
                                Dim theEnum As ExportItem
                                theEnum = ExportTable(props(i).Misc - 1)
                                prop.Value = theEnum.GetEnumItem(num)
                            Else
                                prop.Value = num
                            End If

                        Case EPropType.T_Index
                            prop.Value = GetResName(r.ReadInt32)
                        Case EPropType.T_Name
                            prop.Value = GetName(r.ReadUInt16)
                        Case EPropType.T_String
                            prop.Value = r.ReadCString(props(i).ArraySize)
                        Case EPropType.T_Vector
                            Dim vect = r.ReadVector

                            If FinalSyntax Then
                                prop.Value = vect.ToStructString
                            Else
                                prop.Value = "(" & vect.ToT3DString & ")"
                            End If

                        Case EPropType.T_Rotator
                            Dim rot = r.ReadRotator

                            If FinalSyntax Then
                                prop.Value = rot.ToStructString
                            Else
                                prop.Value = "(" & rot.ToT3DString & ")"
                            End If
                    End Select

                    result.Add(prop)
                Next

            Next

            Return result
        End Function

        Function ActorSize() As Integer
            If PackageVersion <= 9 Then
                Return 1043
            Else
                Return 1148
            End If
        End Function

    End Class

    Public Class ExportItem
        Public MyLevel As Package

        Public Name As String
        Public TypeNum As EResType
        Public HeaderSize As UInt32
        Public HeaderPtr As UInt32
        Public DataPtr As UInt32
        Public DataSize As UInt32

        Function Tab() As String
            Return Space(MyLevel.Indent)
        End Function

        Function GetTexture() As Bitmap
            Dim headerPtr As UInt32 = Me.HeaderPtr
            Dim width, height, paletteNum As Int32
            MyLevel.r.BaseStream.Position = headerPtr + 80
            paletteNum = MyLevel.r.ReadUInt32() - 1 '1-based to 0-based
            MyLevel.r.BaseStream.Position = headerPtr + 160
            width = MyLevel.r.ReadUInt32()
            height = MyLevel.r.ReadUInt32()

            'read palette
            Dim palette(256 - 1) As Color
            MyLevel.r.BaseStream.Position = MyLevel.ExportTable(paletteNum).DataPtr
            For i = 0 To 255
                palette(i) = RGBXToColor(MyLevel.r.ReadInt32)
            Next

            'go to beginning of largest (last) mipmap
            MyLevel.r.BaseStream.Position = Me.DataPtr + Me.DataSize - (width * height)
            Dim bmp As New Bitmap(width, height)
            Dim fastBmp As New FastPixel(bmp)
            fastBmp.Lock()

            For y = 0 To height - 1
                For x = 0 To width - 1
                    fastBmp.SetPixel(x, y, palette(MyLevel.r.ReadByte()))
                Next
            Next

            fastBmp.Unlock(True)

            Return bmp

        End Function

        Function GetHeader() As Byte()
            MyLevel.r.BaseStream.Position = Me.HeaderPtr
            Return MyLevel.r.ReadBytes(Me.HeaderSize)
        End Function

        Function GetData() As Byte()
            MyLevel.r.BaseStream.Position = Me.DataPtr
            Return MyLevel.r.ReadBytes(Me.DataSize)
        End Function

        Function GetSound() As Byte()
            MyLevel.r.BaseStream.Position = Me.DataPtr + 252
            Return MyLevel.r.ReadBytes(Me.DataSize - 252)
        End Function

        Sub SaveMesh(anivfile As String, datafile As String)
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 152
            Dim polygonCount = MyLevel.r.ReadUInt16()
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 160
            Dim vertexCount = MyLevel.r.ReadUInt16()
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 176
            Dim frameCount = MyLevel.r.ReadUInt16()

            Dim dataWriter As New IO.BinaryWriter(New IO.FileStream(datafile, IO.FileMode.CreateNew, IO.FileAccess.ReadWrite))
            dataWriter.Write(CUShort(polygonCount))
            dataWriter.Write(CUShort(vertexCount))
            dataWriter.Seek(44, IO.SeekOrigin.Current)
            MyLevel.r.BaseStream.Position = Me.DataPtr
            dataWriter.Write(MyLevel.r.ReadBytes(16 * polygonCount))
            dataWriter.Close()

            Dim animWriter As New IO.BinaryWriter(New IO.FileStream(anivfile, IO.FileMode.CreateNew, IO.FileAccess.ReadWrite))
            animWriter.Write(CUShort(frameCount))
            animWriter.Write(CUShort(vertexCount * 4))
            animWriter.Write(MyLevel.r.ReadBytes(CInt(frameCount) * CInt(vertexCount) * 4))
            animWriter.Close()

        End Sub

        Function GetText() As String
            MyLevel.r.BaseStream.Position = Me.DataPtr
            Return MyLevel.r.ReadCString(Me.DataSize)
        End Function

        Function GetTextureInfo() As TextureInfo
            Dim result As New TextureInfo

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 80
            result.PaletteNum = MyLevel.r.ReadUInt32() - 1
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 160
            result.Width = MyLevel.r.ReadUInt32()
            result.Height = MyLevel.r.ReadUInt32()
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 170
            result.HasMipmaps = MyLevel.r.ReadByte()

            Return result
        End Function

        Function GetModel() As String
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 76
            Dim sideCount = MyLevel.r.ReadUInt16

            Dim text As New System.Text.StringBuilder

            If MyLevel.FinalSyntax Then
                text.AppendLine(Tab() & "Begin Brush Name=" & Me.Name)
                MyLevel.Indent += 3
            End If

            text.AppendLine(Tab() & "Begin PolyList" & If(MyLevel.FinalSyntax, "", " Num=" & sideCount))
            MyLevel.Indent += 3
            For i = 0 To sideCount - 1
                Dim sidePtr = Me.DataPtr + i * 352
                MyLevel.r.BaseStream.Position = sidePtr

                Dim origin, normal, texU, texV As Vector
                Dim flags, group, item, texture As Integer
                Dim vertCount, uPan, vPan As Byte
                Dim verts As New List(Of Vector)
                Dim texName As String

                origin = MyLevel.r.ReadVector()
                MyLevel.r.ReadUInt32()
                normal = MyLevel.r.ReadVector
                MyLevel.r.ReadUInt32()
                texU = MyLevel.r.ReadVector / 65536
                MyLevel.r.ReadUInt32()
                texV = MyLevel.r.ReadVector / 65536
                MyLevel.r.ReadUInt32()
                MyLevel.r.BaseStream.Position = sidePtr + 336
                vertCount = MyLevel.r.ReadByte
                MyLevel.r.BaseStream.Position = sidePtr + 320
                flags = MyLevel.r.ReadInt32
                MyLevel.r.BaseStream.Position = sidePtr + 328
                texture = MyLevel.r.ReadUInt16
                If texture = 0 Then
                    texName = "DefaultTexture"
                Else
                    texName = MyLevel.ExportTable(texture - 1).Name
                End If
                MyLevel.r.BaseStream.Position = sidePtr + 332
                group = MyLevel.r.ReadUInt16
                item = MyLevel.r.ReadUInt16
                MyLevel.r.BaseStream.Position = sidePtr + 64
                For j = 0 To vertCount - 1
                    verts.Add(MyLevel.r.ReadVector)
                    MyLevel.r.ReadUInt32()
                Next
                MyLevel.r.BaseStream.Position = sidePtr + 348
                uPan = MyLevel.r.ReadByte
                vPan = MyLevel.r.ReadByte

                text.Append(Tab() & "Begin Polygon")
                MyLevel.Indent += 3
                If group <> 65535 Then
                    text.Append(" Group=" & MyLevel.GetName(group))
                End If
                If item <> 65535 Then
                    text.Append(" Item=" & MyLevel.GetName(item))
                End If
                text.Append(" Texture=" & texName)
                If MyLevel.RemapPolyFlags Then
                    flags = RemapPolyFlags(flags)
                End If
                text.AppendLine(" Flags=" & flags)

                text.AppendLine(Tab() & "Origin " & origin.ToT3DString)
                text.AppendLine(Tab() & "Normal " & normal.ToT3DString)
                text.AppendLine(Tab() & "TextureU " & texU.ToT3DString)
                text.AppendLine(Tab() & "TextureV " & texV.ToT3DString)
                text.AppendLine(Tab() & "Pan U=" & uPan & " V=" & vPan)
                For j = 0 To vertCount - 1
                    text.AppendLine(Tab() & "Vertex " & verts(j).ToT3DString)
                Next
                MyLevel.Indent -= 3
                text.AppendLine(Tab() & "End Polygon")

            Next
            MyLevel.Indent -= 3
            text.AppendLine(Tab() & "End PolyList")

            If MyLevel.FinalSyntax Then
                MyLevel.Indent -= 3
                text.AppendLine(Tab() & "End Brush")
            End If

            Return text.ToString
        End Function

        Function GetBrushModel() As ExportItem
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 96
            Return MyLevel.ExportTable(MyLevel.r.ReadUInt16 - 1)
        End Function

        Function GetMissingMoverProps() As String
            Dim text As New System.Text.StringBuilder

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 140
            Dim prepivot = MyLevel.r.ReadVector

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 156
            Dim scale = MyLevel.r.ReadVector
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 172
            Dim sheerrate = MyLevel.r.ReadSingle
            Dim sheeraxis = MyLevel.r.ReadByte

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 216
            Dim postscale = MyLevel.r.ReadVector
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 232
            Dim postsheerrate = MyLevel.r.ReadSingle
            Dim postsheeraxis = MyLevel.r.ReadByte

            text.AppendLine(Tab() & "PrePivot=" & prepivot.ToStructString)
            text.AppendLine(Tab() & "MainScale=(Scale=" & scale.ToStructString & ",SheerRate=" &
                                NumToStr(sheerrate) & ",SheerAxis=" & sheeraxis & ")")
            text.AppendLine(Tab() & "PostScale=(Scale=" & postscale.ToStructString & ",SheerRate=" &
                            NumToStr(postsheerrate) & ",SheerAxis=" & postsheeraxis & ")")

            Return text.ToString
        End Function

        Function GetBrush() As String
            Dim model As UInt16
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 96
            model = MyLevel.r.ReadUInt16 - 1

            Dim text As New System.Text.StringBuilder

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 200
            Dim postpivot = MyLevel.r.ReadVector
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 112
            Dim location = MyLevel.r.ReadVector

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 128
            Dim rotation = MyLevel.r.ReadRotator
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 140
            Dim prepivot = MyLevel.r.ReadVector
            
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 156
            Dim scale = MyLevel.r.ReadVector
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 172
            Dim sheerrate = MyLevel.r.ReadSingle
            Dim sheeraxis = MyLevel.r.ReadByte

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 216
            Dim postscale = MyLevel.r.ReadVector
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 232
            Dim postsheerrate = MyLevel.r.ReadSingle
            Dim postsheeraxis = MyLevel.r.ReadByte

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 180
            Dim csgOper = MyLevel.r.ReadByte

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 192
            Dim flags = MyLevel.r.ReadInt32
            If MyLevel.RemapPolyFlags Then
                flags = RemapPolyFlags(flags)
            End If

            text.AppendLine(Tab() & "Begin " & If(MyLevel.FinalSyntax, "Actor Class=Brush", "Brush") & " Name=" & Me.Name)

            MyLevel.Indent += 3

            If MyLevel.FinalSyntax Then
                text.AppendLine(Tab() & "Location=" & (location + postpivot).ToStructString)
                text.AppendLine(Tab() & "Rotation=" & rotation.ToStructString)
                text.AppendLine(Tab() & "PrePivot=" & prepivot.ToStructString)
                text.AppendLine(Tab() & "MainScale=(Scale=" & scale.ToStructString & ",SheerRate=" &
                                NumToStr(sheerrate) & ",SheerAxis=" & sheeraxis & ")")
                text.AppendLine(Tab() & "PostScale=(Scale=" & postscale.ToStructString & ",SheerRate=" &
                                NumToStr(postsheerrate) & ",SheerAxis=" & postsheeraxis & ")")
                text.AppendLine(Tab() & "CsgOper=" & csgOper)
                text.AppendLine(Tab() & "PolyFlags=" & flags)
            Else
                text.AppendLine(Tab() & "Location " & location.ToT3DString)
                text.AppendLine(Tab() & "Rotation " & rotation.ToT3DString)
                text.AppendLine(Tab() & "PrePivot " & prepivot.ToT3DString)
                text.AppendLine(Tab() & "PostPivot " & postpivot.ToT3DString)
                text.AppendLine(Tab() & "Scale " & scale.ToOldT3DString & " S=" & NumToStr(sheerrate) &
                                " AXIS=" & sheeraxis)
                text.AppendLine(Tab() & "PostScale " & postscale.ToOldT3DString & " S=" & NumToStr(postsheerrate) &
                                " AXIS=" & postsheeraxis)
                text.AppendLine(Tab() & "Settings CSG=" & csgOper & " PolyFlags=" & flags)
            End If

            text.Append(MyLevel.ExportTable(model).GetModel)

            If MyLevel.FinalSyntax Then
                text.AppendLine(Tab() & "Brush=Model'" & MyLevel.ExportTable(model).Name & "'")
            End If

            MyLevel.Indent -= 3

            text.AppendLine(Tab() & "End " & If(MyLevel.FinalSyntax, "Actor", "Brush"))

            Return text.ToString

        End Function

        Function GetBrushList() As Integer()
            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 76
            Dim brushCount = MyLevel.r.ReadUInt16

            MyLevel.r.BaseStream.Position = Me.DataPtr
            Dim brushList(brushCount - 1) As Integer

            For i = 0 To brushCount - 1
                brushList(i) = MyLevel.r.ReadUInt32 - 1
            Next

            Return brushList
        End Function

        Function GetClassProperties() As ClassProperty()
            If Me.TypeNum <> EResType.RES_Class Then
                Throw New Exception("GetClassProperties: Not a class")
            End If

            MyLevel.r.BaseStream.Position = Me.HeaderPtr + 76
            Dim propCount = MyLevel.r.ReadUInt16

            MyLevel.r.BaseStream.Position = Me.DataPtr
            Dim props(propCount - 1) As ClassProperty
            For i = 0 To propCount - 1
                props(i).Name = MyLevel.GetName(MyLevel.r.ReadUInt16)
                Dim category = MyLevel.r.ReadUInt16
                If category = 65535 Then
                    props(i).Category = "None"
                Else
                    props(i).Category = MyLevel.GetName(category)
                End If
                props(i).Type = MyLevel.r.ReadUInt32
                props(i).Offset = MyLevel.r.ReadUInt32
                props(i).Flags = MyLevel.r.ReadUInt32
                props(i).ArraySize = MyLevel.r.ReadUInt32
                props(i).ElementSize = MyLevel.r.ReadUInt32
                props(i).TotalSize = MyLevel.r.ReadUInt32
                props(i).Misc = MyLevel.r.ReadInt32
            Next

            Return props
        End Function

        Function GetActor(num As Integer, props As ClassProperty()) As List(Of ActorProperty)
            Return MyLevel.ReadActor(Me.DataPtr + num * MyLevel.ActorSize, props)
        End Function

        Function GetClassDefaults(props As ClassProperty()) As List(Of ActorProperty)
            Return MyLevel.ReadActor(Me.HeaderPtr + 124, props)
        End Function

        Function GetEnumItem(num As Byte) As String
            MyLevel.r.BaseStream.Position = Me.DataPtr + num * 2
            Return MyLevel.GetName(MyLevel.r.ReadUInt16)
        End Function

    End Class

    Structure TextureInfo
        Dim Width As Integer
        Dim Height As Integer
        Dim HasMipmaps As Boolean
        Dim PaletteNum As Integer
    End Structure

    Enum EPropType
        T_None
        T_Byte
        T_Int
        T_Bool
        T_Real
        T_Actor
        T_Index
        T_Name
        T_String
        T_Vector
        T_Rotator
    End Enum

    Const CPF_Edit = &H1    ' Property is user-settable in the editor
    Const CPF_EditInPlay = &H2    ' Property can safely be edited during gameplay
    Const CPF_Const = &H4    ' Actor's property always matches class's default actor property
    Const CPF_Private = &H8    ' Private, inaccessible to scripts
    Const CPF_NoSaveResource = &H10    ' Don't save this CPT_Resource on disk
    Const CPF_ExportResource = &H20    ' Resource can be exported with actor
    Const CPF_Param = &H40    ' Function/When call parameter
    Const CPF_OptionalParam = &H80    ' Optional parameter (if CPF_Param is set)
    Const CPF_ReturnValue = &H100    ' Return value
    Const CPF_Initialized = &H200    ' UnrealScript has specified an initializer value

    Function ParsePropFlags(flags As Integer) As String
        Dim result = ""
        If flags And CPF_Edit Then result &= "Edit"
        If flags And CPF_EditInPlay Then result &= " EditInPlay"
        If flags And CPF_Const Then result &= " Const"
        If flags And CPF_Private Then result &= " Private"
        If flags And CPF_NoSaveResource Then result &= " NoSave"
        If flags And CPF_ExportResource Then result &= " Export"
        If flags And CPF_Param Then result &= " Param"
        If flags And CPF_OptionalParam Then result &= " Optional"
        If flags And CPF_ReturnValue Then result &= " ReturnValue"
        If flags And CPF_Initialized Then result &= " Initialized"
        Return result
    End Function

    Structure ClassProperty
        Dim Name As String
        Dim Category As String
        Dim Type As EPropType
        Dim Offset As Integer
        Dim Flags As Integer
        Dim ArraySize As Integer
        Dim ElementSize As Integer
        Dim TotalSize As Integer
        Dim Misc As Integer
    End Structure

    Enum EResType
        RES_None = 0         ' No resource
        RES_Buffer = 1       ' A large binary object holding misc saveable data
        RES_Array = 3        ' An array of resources
        RES_TextBuffer = 4   ' A text buffer
        RES_Texture = 5      ' Texture or compound texture
        RES_Font = 6         ' Font for use in game
        RES_Palette = 7      ' A palette
        RES_Script = 9       ' Script
        RES_Class = 10       ' An actor class
        RES_ActorList = 11   ' An array of actors
        RES_Sound = 12       ' Sound effect
        RES_Mesh = 14        ' Animated mesh
        RES_Vectors = 16     ' 32-bit floating point vector list
        RES_BspNodes = 17    ' Bsp node list
        RES_BspSurfs = 18    ' Bsp polygon list
        RES_LightMesh = 19   ' Bsp polygon lighting mesh
        RES_Polys = 20       ' Editor polygon list
        RES_Model = 21       ' Model or level map
        RES_Level = 22       ' A game level
        RES_Camera = 25      ' A rendering camera on this machine
        RES_Player = 28      ' A remote player logged into the local server
        RES_VertPool = 29    ' A vertex pool corresponding to a Bsp and FPoints/FVectors table
        RES_Ambient = 30     ' An ambient sound definition
        RES_TransBuffer = 31 ' Transaction tracking buffer
        RES_MeshMap = 32     ' MeshMap
        RES_Bounds = 33      ' Bounding structure
        RES_Terrain = 34     ' Terrain
        RES_Enum = 35        ' Enumeration (array of FName's)
    End Enum

    Public Function TypeToName(typeID As EResType) As String
        Return typeID.ToString.Substring(4) & " (" & typeID & ")"
    End Function

    Structure ActorProperty
        Dim Name As String
        Dim Value As String
    End Structure
End Module
