﻿Public Module Utils
    Public Class UnBinaryReader
        Inherits IO.BinaryReader

        Sub New(input As IO.Stream)
            MyBase.New(input)
        End Sub

        Sub New(file As String)
            Me.New(New IO.FileStream(file, IO.FileMode.Open, IO.FileAccess.Read))
        End Sub

        Public Function ReadCString(Optional ByVal maxLen As UInteger = 0) As String
            Dim str As New System.Text.StringBuilder
            Dim len As UInteger = 0
            Dim curchr As Byte

            Do
                curchr = Me.ReadByte
                len += 1
                If curchr = 0 Then
                    Exit Do
                End If
                str.Append(Chr(curchr))
            Loop Until (len >= maxLen And maxLen > 0)

            Return str.ToString
        End Function

        Function ReadVector() As Vector
            Dim x, y, z As Single
            x = Me.ReadSingle
            y = Me.ReadSingle
            z = Me.ReadSingle
            Return New Vector(x, y, z)
        End Function

        Function ReadRotator() As Rotator
            Dim pitch, yaw, roll As Integer
            pitch = Me.ReadInt32
            yaw = Me.ReadInt32
            roll = Me.ReadInt32
            Return New Rotator With {.Pitch = pitch, .Yaw = yaw, .Roll = roll}
        End Function


    End Class

    Public Function RGBXToColor(num As Integer) As Color
        Return Color.FromArgb(num And 255, (num >> 8) And 255, (num >> 16) And 255)
    End Function

    Function NumToStr(ByVal val As Double) As String
        'don't use period and "e" notation
        Return val.ToString("F6", Globalization.CultureInfo.InvariantCulture)
    End Function

    Function StrToNum(val As String) As Single
        Return Single.Parse(val, Globalization.CultureInfo.InvariantCulture)
    End Function


    Public Structure Vector
        Dim X As Single
        Dim Y As Single
        Dim Z As Single

        Sub New(_x As Single, _y As Single, _z As Single)
            Me.X = _x
            Me.Y = _y
            Me.Z = _z
        End Sub

        Function ToStructString() As String
            Return "(X=" & NumToStr(X) & ",Y=" & NumToStr(Y) & ",Z=" & NumToStr(Z) & ")"
        End Function

        Function ToOldT3DString() As String
            Return "X=" & NumToStr(X) & " Y=" & NumToStr(Y) & " Z=" & NumToStr(Z)
        End Function

        Function ToT3DString() As String
            Return NumToStr(X) & "," & NumToStr(Y) & "," & NumToStr(Z)
        End Function

        Shared Operator +(lhs As Vector, rhs As Vector) As Vector
            Return New Vector(lhs.X + rhs.X, lhs.Y + rhs.Y, lhs.Z + rhs.Z)
        End Operator

        Shared Operator -(lhs As Vector, rhs As Vector) As Vector
            Return New Vector(lhs.X - rhs.X, lhs.Y - rhs.Y, lhs.Z - rhs.Z)
        End Operator

        Shared Operator *(lhs As Vector, rhs As Double) As Vector
            Return New Vector(lhs.X * rhs, lhs.Y * rhs, lhs.Z * rhs)
        End Operator

        Shared Operator *(lhs As Double, rhs As Vector) As Vector
            Return New Vector(rhs.X * lhs, rhs.Y * lhs, rhs.Z * lhs)
        End Operator

        Shared Operator /(lhs As Vector, rhs As Double) As Vector
            Return New Vector(lhs.X / rhs, lhs.Y / rhs, lhs.Z / rhs)
        End Operator
    End Structure

    Public Structure Rotator
        Dim Pitch As Integer
        Dim Yaw As Integer
        Dim Roll As Integer

        Function ToStructString() As String
            Return "(Pitch=" & Pitch & ",Yaw=" & Yaw & ",Roll=" & Roll & ")"
        End Function

        Function ToT3DString() As String
            Return Pitch & "," & Yaw & "," & Roll
        End Function
    End Structure


End Module
